MICRO_TO_MILLI = 1e-3
MILLI_TO_MICRO = 1e3
DEGREES_TO_SECONDS = 3600.0
SECONDS_TO_DEGREES = 1 / 3600.0
INCHES_TO_MILLIMETERS = 25.4
PERCENT_TO_DECIMAL = 0.01
FWHM_TO_SIGMA = 1 / 2.355
BASE_TO_KILO = 1e-3
KILO_TO_BASE = 1e3
SUN_MAGNITUDE = -26.74
ZERO = 1e-12
MAD_TO_STD = 1.4826
